{ config, lib, inputs, pkgs, options, specialArgs, ... }:

let
  var = options.variable;

  my-lib = import ./lib/default.nix { inherit lib; };

  # (k: k + k) -> { a = 1; } -> { aa = 1; }
  mapKeys = f: lib.mapAttrs' (k: v: lib.nameValuePair (f k) v);

  # (v: 2 * v) -> { a = 1; } -> { a = 2; }
  mapVals = f: lib.mapAttrs (_: f);

  # { b = 0; } -> { c = { a = 1; } } -> { c = { b = 0; a = 1; } }
  default = defaults: mapVals (v: defaults // v);

  # "b" -> { a = 1; } -> { b_a = 1; }
  inNamespace = prefix: mapKeys (k: "${prefix}_${k}");

  # { a = 1; } -> { name = "a"; a = 1; }
  setNames = lib.mapAttrs (k: v: { name = k; } // v);

  # "foo" -> "\${data.sops_file.secrets.data[\"foo\"]}"
  secret = str: lib.tfRef "data.sops_file.secrets.data[\"${str}\"]";

  environment = lib.tfRef "var.environment";

  hetzner = let

    # https://docs.hetzner.com/cloud/api/getting-started/generating-api-token
    token = secret "hcloud_api_token";

    # hetzner cloud zones, as attrset cuz less fragile than strings
    # { eu-central = "eu-central"; ... }
    zones = lib.listToAttrs (lib.lists.map (k: { name = k; value = k; })) [
      "eu-central"
      "us-east"
      "us-west"
    ];

    # https://docs.hetzner.com/cloud/general/locations/
    # { nbg1 = "eu-central"; ... }
    locationZone = with zones; {
      fsn1 = eu-central;  # DE Falkenstein
      nbg1 = eu-central;  # DE Nuremberg
      hel1 = eu-central;  # FI Helsinki
      ash = us-east;  # US Ashburn, VA
      hil = us-west;  # US Hillsboro, OR
    };

    # https://registry.terraform.io/providers/hetznercloud/hcloud/latest/docs#delete-protection
    # The Hetzner Cloud API allows to protect resources from deletion by putting a "lock" on them.
    # This can also be configured through OpenTofu through the `delete_protection` argument on resources that support it.
    # Please note, that this does not protect deletion from OpenTofu itself,
    # as the Provider will lift the lock in that case.
    # If you also want to protect your resources from deletion by OpenTofu,
    # you can use the [`prevent_destroy` lifecycle attribute](https://opentofu.org/docs/language/meta-arguments/lifecycle#syntax-and-arguments).
    delete_protection = true;

    # Whether auto delete is enabled.
    # `Important note:` It is recommended to set `auto_delete` to `false`,
    # because if a server assigned to a managed ip is getting deleted,
    # it will also delete the primary IP which will break the TF state.
    auto_delete = false;

    # https://docs.hetzner.com/cloud/servers/overview/#pricing
    # cheapest shared-vcpu arm type
    serverType = "cax11";

    # https://docs.hetzner.com/cloud/general/locations/
    location = lib.tfRef "var.hcloud_location";
    network_zone = locationZone.${location};

    labels = {
      inherit environment;
    };

  in { inherit token delete_protection auto_delete location network_zone serverType labels; };

  backend = {

    local = {};

    # cloud = {
    #   hostname = "app.terraform.io";
    #   organization = "bij1";
    #   workspaces = {
    #     name = "hcloud";
    #   };
    # };

  };

in rec {

  terraform = {
    
    required_providers = {

      sops.source = "carlpett/sops";

    };

    cloud = {
      hostname = "app.terraform.io";
      organization = "bij1";
      workspaces = {
        name = "hcloud";
      };
    };

  };

  provider = {

    sops = {};

    # Configure the Hetzner Cloud Provider
    hcloud.token = lib.mkForce hetzner.token;

  };

  resource = (inNamespace "hcloud" {

    ssh_key = setNames (lib.mapAttrs (_: v: { public_key = v; }) my-lib.ssh-keys);

    # network = setNames {

    #   production-net = {
    #     ip_range = "10.0.0.0/24";
    #   };

    # };

    # network_route = {

    #   production-route = {
    #     network_id  = lib.tfRef "hcloud_network.production-net.id";
    #     destination = "10.100.1.0/24";
    #     gateway     = "10.0.1.1";
    #   };

    # };

    # network_subnet = mapVals (default { inherit (hetzner) network_zone; }) {

    #   production-subnet = {
    #     network_id   = lib.tfRef "hcloud_network.production-net.id";
    #     type         = "cloud";
    #     ip_range     = "10.0.1.0/24";
    #   };

    # };

    # placement_group = setNames (mapVals (default { type = "spread"; }) {

    #   production-dbs = {
    #     labels = {
    #       environment = "production";
    #     };
    #   };

    # });

    # firewall = setNames {

    #   production-firewall = {

    #     rule = {
    #       # direction = "in";
    #       # protocol  = "icmp";
    #       # source_ips = [
    #       #   "0.0.0.0/0"
    #       #   "::/0"
    #       # ];
    #     };

    #     rule = {
    #       # direction = "in";
    #       # protocol  = "tcp";
    #       # port      = "80-85";
    #       # source_ips = [
    #       #   "0.0.0.0/0"
    #       #   "::/0"
    #       # ];
    #     };

    #   };

    # };

    # managed_certificate = setNames (mapVals (default { inherit (hetzner) labels; }) {

    #   production-cert-wordpress-main = {
    #     domain_names = ["*.bij1.org" "bij1.org"];
    #   };

    # });

    # primary_ip = setNames (mapVals (default { inherit (hetzner) delete_protection auto_delete labels; }) {

    #   production-ip-nextcloud = {
    #     datacenter    = "nbg1-dc6";  # https://docs.hetzner.com/general/others/data-centers-and-connection/
    #     type          = "ipv6";
    #     assignee_type = "server";
    #   };

    # });

    # floating_ip = setNames (mapVals (default { home_location = hetzner.location; }) {

    #   production-ip-nextcloud = {
    #     type = "ipv6";
    #   };

    # });

  }) // {

    local_file.test_import = {
      filename = "test_import.txt";
      content = config.sops.secrets.tf_cloud_token.path;
      # content = "lol";
    };

  };

  # Set the variable value in *.tfvars file
  # or using -var="hcloud_api_token=..." CLI option
  # given e.g. variable.foo, use as `lib.tfRef "var.foo"`
  variable = {

    hcloud_api_token = {
      type = "string";
      description = "[Hetzner Cloud API Token](https://docs.hetzner.com/cloud/api/getting-started/generating-api-token)";
      default = hetzner.token;
      sensitive = true;
    };

    environment = {
      type = "string";
      description = "the environment to deploy to, e.g. production, staging";
      default = "production";
    };

    hcloud_location = {
      type = "string";
      description = "[Hetzner Cloud Location](https://docs.hetzner.com/cloud/general/locations/)";
      default = "nbg1";
    };

  };

  # !!! `provisioner` is impure
  # configure provisioning private Key to be used when running provisioning on the machines
  # `file` and `remote-exec` steps of `provisioners` automatically get the connection part setup,
  # if you configure the `provisioner.privateKeyFile` parameter.
  # provisioner.privateKeyFile = toString ./sshkey;

  # https://github.com/terranix/terranix-hcloud/blob/main/options.md
  hcloud = {
    enable = true;
    # can also be specified with the TF_VAR_hcloud_api_token environment variable
    provider = { inherit (hetzner) token; };
    export.nix = "hetzner.nix";

    # https://github.com/terranix/terranix-examples/blob/main/hcloud-nixos-server/config.nix#L18
    # https://nix-community.github.io/srvos/getting_started/
    server = setNames (mapVals (default { inherit (hetzner) location serverType; enable = true; }) {
      # ssh root@$( tofu output nixserver-server1_ipv4_address ) -i ./sshkey
      # server1 = {
      #   # common options
      #   backups = false;  # billed +20%
      #   extraConfig = {};
      #   # provisioners = [];
      #   # provisioners = [{"file":{"destination":"/root/.zshrc","source":"~/.zshrc"}},{"remote-exec":{"inline":["shutdown -r +1"]}}];

      #   # # container-specific config
      #   # image = "ubuntu-22.04";
      # };
    });

    nixserver = setNames (mapVals (default { inherit (hetzner) location serverType; enable = true; channel = "nixos-23.11"; }) {
      # server1 = {
      #   # common options
      #   backups = false;  # billed +20%
      #   extraConfig = {};
      #   # provisioners = [];

      #   # # nix-specific config
      #   # configurationFile = ./my-cfg.nix;
      #   # configurationFile = pkgs.writeText "configuration.nix" ''
      #   #   { pkgs, lib, config, ... }:
      #   #   {
      #   #     environment.systemPackages = [ pkgs.git ];
      #   #   }
      #   # '';
      # };
    });
  };

  data = {

    sops_file.secrets = {
      source_file = "secrets.enc.yaml";
    };

    hcloud_ssh_keys."all_keys" = {};

  };

  output = {

    "keys_output" = {
      value = lib.tfRef "data.hcloud_ssh_keys.all_keys";
    };

  };

}
